
from __future__ import division, print_function

import json

from csv import reader as csv_reader
from operator import itemgetter
from os.path import abspath, basename, dirname, exists, join, splitext
from warnings import warn

from BioExt import AmbigList, translate_ambiguous


__all__ = []


_scoredata = {
    'PI': {'name': 'PI.dat', 'offset': 0, 'gene': 'PR'},
    'NRTI': {'name': 'NRTI.dat', 'offset' : 99, 'gene': 'RT'},
    'NNRTI': {'name': 'NNRTI.dat', 'offset' : 99, 'gene': 'RT'}
}

COMPLETE_LIST, BINARY  = range(2)


class DrugCoord(object):
    DELETION, INSERTION, VALUE = range(3)

    def __init__(self, coord, klass, amino, value):
        if klass not in (DrugCoord.DELETION, DrugCoord.INSERTION, DrugCoord.VALUE):
            raise ValueError('DrugCoord klass must be one of DrugCoord.{DELETION, INSERTION, or VALUE}')
        if not isinstance(coord, int):
            raise ValueError('DrugCoord coord must be an index into a sequence')

        self._coord = coord
        self._klass = klass
        self._amino = amino.upper()
        self._value = value

    def __call__(self, seq):
        if not isinstance(seq, AmbigList):
            raise ValueError('seq must be an AmbigList, generated by translate_ambiguous in BioExt')

        c = self._coord
        if c >= len(seq):
            warn("sequence too short to evaluate at position '%d'" % self._coord)
            return c, 0, 0

        ndegen = len(seq[c])
        r = 0
        if self._klass == DrugCoord.DELETION:
            if seq[c] == '-':
                r = self._value
        elif self._klass == DrugCoord.INSERTION:
            warn('INSERTION UNEVALUATED!!! We need to think about how to communicate insertion events')
        else:
            if self._amino in seq[c]:
                r = self._value
        return c, r, ndegen


class DrugData(object):

    def __init__(self, name, gene, drug_coords):
        self._name = name
        self._gene = gene
        self._drug_coords = drug_coords

    def __call__(self, seq):
        # sort by the coordinate
        tmp = sorted((d(seq) for d in self._drug_coords), key=itemgetter(0))
        # group by the coordinate
        r = []
        prev = -1
        for coord, s, n in tmp:
            # if we're the previous coord,
            # append to that list
            if len(r) > 0 and coord == prev:
                r[-1].append((coord, s, n))
            # otherwise, make a new list
            else:
                r.append([(coord, s, n)])
            prev = coord
        return r


def read_scorefile(filepath, gene, offset):
    if not exists(filepath):
        raise ValueError("file '%s' doesn't exist!" % filepath)

    with open(filepath) as fh:
        reader = csv_reader(fh, delimiter='\t')

        header = reader.next()

        # some error correction
        if len(header) < 3:
            raise RuntimeError('expected at least 3 columns')
        if header[0].lower() != 'position':
            raise RuntimeError("expected first column label to be 'position'")
        if header[1].lower() != 'aa':
            raise RuntimeError("expected second column label to be 'aa'")

        drug_idxs = [(i + 2, d) for i, d in enumerate(header[2:])]
        pos_by_drug = dict((d, []) for d in header[2:])

        for row in reader:
            # coordinates are assumed 1-indexed
            coord = int(row[0]) + offset - 1
            amino = row[1].upper()
            klass = (DrugCoord.DELETION  if amino == '-' else
                     DrugCoord.INSERTION if amino == 'INSERT' else
                     DrugCoord.VALUE)
            for idx, drug in drug_idxs:
                val = int(row[idx])
                if val > 0:
                    pos_by_drug[drug].append(DrugCoord(coord, klass, amino, val))

        return dict((d, DrugData(d, gene, pos_by_drug[d])) for _, d in drug_idxs)


def load_scoredata():
    scorefiledir = join(dirname(abspath(__file__)), 'data', 'scores')
    # scorefiles = iglob(join(scorefiledir, '*.dat'))
    # return a dict of dicts, key = drug class, value = dict of DrugData's,
    # keyed on the drug name
    r = {}
    for klass, data in _scoredata.items():
        fp = join(scorefiledir, data['name'])
        gene = data['gene']
        off = data['offset']
        r[klass] = read_scorefile(fp, gene, off)
    return r


OPTIMISTIC, PESSIMISTIC = range(2)

def DRAM_calculator_p(msa, mode=None, noise_threshold=4):
    if mode is None:
        mode = OPTIMISTIC

    if mode not in (OPTIMISTIC, PESSIMISTIC):
        raise ValueError('mode must be one of either OPTIMISTIC or PESSIMISTIC')

    table = load_scoredata()

    resolver = max if mode == OPTIMISTIC else min

    r = []
    for seq in msa:
        ambiglist = translate_ambiguous(seq)
        seq_r = {}
        for klass, drugdatas in table.items():
            klass_r = {}
            for drug, drugdata in drugdatas.items():
                # grab a list of lists,
                # where each sublist is for a particular coord
                coord_groups = drugdata(ambiglist)
                # now grab either the min or max from these sublists,
                # depending on how optimistic or pessimistic we are being
                # and use a generator as we can calculate in a single pass
                coord_score_ndegens = (resolver(g, key=itemgetter(1)) for g in coord_groups)
                # grab some stats, in a single pass
                covered, missing, noise, score = 0, 0, 0, 0
                for _, s, n in coord_score_ndegens:
                    if n > noise_threshold:
                        noise += 1
                    elif s > 0:
                        covered += 1
                        score += s
                    else: # s == 0 => missing
                        missing += 1
                klass_r[drug] = (covered, missing, noise, score)
            seq_r[klass] = klass_r
        r.append(seq_r)
    return r


# DRAM_AMBIGS_PESSIMISTIC, DRAM_AMBIGS_OPTIMISTIC = range (2)
# DRAM_REPORT_COMPLETE, DRAM_REPORT_BY_DRUG, DRAM_REPORT_BY_MUTATION, DRAM_REPORT_BY_CLASS = range (4)
# 
# def DRAM_reporter (positional_resistance_data, resistance_table = None, ambiguity_resolution = None, report_mode = None, mut_score_limit = 30, drug_resistance_limit = 30):
# # positional_resistance_data is returned by DRAM_calculator
#     if resistance_table is None:
#         if '>resistance table' in positional_resistance_data:
#             resistance_table = positional_resistance_data['>resistance table']
#         else:
#             resistance_table = import_mdr_scores_from_stanford ()
#     if ambiguity_resolution is None:
#         ambiguity_resolution = DRAM_AMBIGS_PESSIMISTIC
#     if report_mode is None:
#         report_mode = DRAM_REPORT_COMPLETE
# 
#     dram_report = {}
# 
#     # generate the full report by default, then condense if necessary
#     # score by drug (with comments if sequenced completely or has noisy data): min and max
#     # resistance mutations by drug class (can be limited to major)
#     # resistance to classes e.g. NRTI : ['susceptible', 'resistant']
# 
# 
#     for sequence_id in positional_resistance_data:
#         if sequence_id == '>resistance table':
#             continue
#         sequence_info = positional_resistance_data[sequence_id]
#         sequence_report = {'by drug' : {}}
#         for drug_name in resistance_table['by drug']:
#             total_score = 0
#             covered_positions = 0
#             noisy_positions   = 0
#             missing_positions = 0
#             major_DRAM        = []
# 
#             list_of_positions_by_drug = resistance_table['by drug'][drug_name]
# 
#             for position in list_of_positions_by_drug:
#                 score = 0
#                 in_this_sequence = sequence_info[str(position)]
#                 resolution_scores = {}
# 
#                 if in_this_sequence is None:
#                     missing_positions += 1
#                     continue
#                 elif in_this_sequence[0] == 'Noisy':
#                     noisy_positions += 1
#                 else:
#                     covered_positions += 1
#                     for value in in_this_sequence:
#                         if value != 'WT':
#                             try:
#                                 score = resistance_table['mutations'][value][drug_name]
#                             except KeyError:
#                                 score = 0
#                         else:
#                             score = 0
#                         resolution_scores [value] = score
#                     mutation,score  = _DRAM_resolve_score (resolution_scores, ambiguity_resolution)
#                     if score >= mut_score_limit:
#                         major_DRAM.append(resistance_table["mutations"][mutation]["Original"])
# 
#                 total_score += score
# 
#             sequence_report ['by drug'][drug_name] = {'covered' : covered_positions, 'missing' : missing_positions, 'noisy': noisy_positions, 'score':  total_score ,'mutations' : major_DRAM if len (major_DRAM) else None}
# 
# 
#         sequence_report ['by class'] = {}
#         for drug_class, class_info in resistance_table['by class'].items():
#             report = {'mutations' : set(), 'resistant' : []}
#             for drug_id in class_info['names']:
#                 #print (drug_id, report)
#                 if sequence_report ['by drug'][drug_id]['score'] >= drug_resistance_limit:
#                     report['resistant'].append (drug_id)
#                 if sequence_report ['by drug'][drug_id]['mutations'] is not None:
#                     report['mutations'].update (sequence_report ['by drug'][drug_id]['mutations'])
#             report['mutations'] = list (report['mutations'])
#             sequence_report ['by class'][drug_class] = report
# 
#         dram_report [sequence_id] = sequence_report
#         #assert False
# 
#     if report_mode == DRAM_REPORT_BY_DRUG:
#     # in this mode the report is of the form
#     # seq id => list of drugs to which the sequence is resistant, grouped by class
#         drug_report = {}
#         for seq_id in dram_report:
#             seq_drug_report = {}
#             for drug_class, drug_value in dram_report[seq_id]['by class'].items():
#                 seq_drug_report [drug_class] = drug_value['resistant']
#             drug_report [seq_id] = seq_drug_report
#         return drug_report
#     elif report_mode == DRAM_REPORT_BY_MUTATION:
#     # in this mode the report is of the form
#     # seq id => list of major mutations
#         mut_report = {}
#         for seq_id in dram_report:
#             seq_mut_report = set()
#             for drug_class, drug_value in dram_report[seq_id]['by class'].items():
#                 seq_mut_report.update(drug_value['mutations'])
#             mut_report [seq_id] = list(seq_mut_report)
#         return mut_report
#     elif report_mode == DRAM_REPORT_BY_CLASS:
#     # in this mode the report is of the form
#     # seq id => list of drug classes to which there are major mutations
#         class_report = {}
#         for seq_id in dram_report:
#             class_list = set()
#             for drug_class, drug_value in dram_report[seq_id]['by class'].items():
#                 if len (drug_value['resistant']):
#                     class_list.add (drug_class)
#             class_report [seq_id] = list(class_list)
#         return class_report
# 
# 
#     return dram_report


if __name__ == "__main__":
    from sys import argv
    assert len (argv) > 1, "This program requires a single argument -- the .json file of sequences split by positions"

    mode = None

#     if len (argv) == 2:
#         with open (argv[1]) as positional_output:
#             DRAM_data = DRAM_calculator (json.load (positional_output))
#     else:
#         with open (argv[1]) as positional_output:
#             DRAM_data = json.load (positional_output)
#             mode = int (argv[2])
# 
#     print (json.dumps(DRAM_reporter (DRAM_data,report_mode = mode),sort_keys=True, indent=4))





__version__ = '0.4.3'

from ._annotate import *
from ._codonaligner import *
from ._common import *
from ._compensatory import *
from ._divestimator import *
from ._fel import *
from ._mdr_variants import *
from ._preprocessing import *
from ._rate_class_neb import *
from ._region_extract import *
from ._reporter import *
from ._variants import *


__all__ = []
__all__ += _annotate.__all__
__all__ += _codonaligner.__all__
__all__ += _common.__all__
__all__ += _compensatory.__all__
__all__ += _divestimator.__all__
__all__ += _fel.__all__
__all__ += _mdr_variants.__all__
__all__ += _preprocessing.__all__
__all__ += _rate_class_neb.__all__
__all__ += _region_extract.__all__
__all__ += _reporter.__all__
__all__ += _variants.__all__


try:
    from ._graph import *
    __all__ += _graph.__all__
except ImportError:
    from warnings import warn
    warn('graphing disabled because some required modules are missing')




from __future__ import division, print_function

from os.path import exists

import numpy as np

from matplotlib.path import Path
from matplotlib.patches import PathPatch

from freetype import *


class Basefont(object):

    def __init__(self, filepath, charsize=48):
        if not exists(filepath):
            raise ValueError('No valid font file specified')

        try:
            if isinstance(filepath, unicode):
                filepath = filepath.encode('utf-8')
        except NameError:
            if not isinstance(filepath, bytes):
                filepath = filepath.encode('utf-8')

        face = Face(filepath)
        face.set_char_size(charsize * 64)
        self.face = face
        self.cache = {}

    def __getitem__(self, value):
        return Basefont.char_patch(self, value)

    def char_patch(self, char):
        # if we have the path in our cache, re-use it
        if char in self.cache:
            return PathPatch(self.cache[char])

        self.face.load_char(char)
        slot = self.face.glyph
        outline = slot.outline

        points_ = np.array(outline.points, dtype=[('x', float), ('y', float)])
        xs, ys = points_['x'], points_['y']

        xmin, ymin = xs.min(), ys.min()
        dx, dy = (xs.max() - xmin), (ys.max() - ymin)

        start, end = 0, 0
        verts, codes = [], []

        for i in range(len(outline.contours)):
            end = outline.contours[i]
            points = outline.points[start:end+1]
            points.append(points[0])
            # normalize points to (0, 1) space
            points = [((x - xmin) / dx, (y - ymin) / dy) for x, y in points]
            tags = outline.tags[start:end+1]
            tags.append(tags[0])

            verts.append(points[0])
            codes.append(Path.MOVETO)

            segments = [[points[0]]]
            for j in range(1, len(points)):
                segments[-1].append(points[j])
                if tags[j] & (1 << 0) and j < (len(points) - 1):
                    segments.append([points[j]])

            for segment in segments:
                if len(segment) == 2:
                    verts.extend(segment[1:])
                    codes.append(Path.LINETO)
                elif len(segment) == 3:
                    verts.extend(segments[1:])
                    codes.extend([Path.CURVE3, Path.CURVE3])
                else:
                    verts.append(segment[1])
                    codes.append(Path.CURVE3)
                    for i in range(1, len(segment)-2):
                        A, B = segment[i], segment[i + 1]
                        C = ((A[0] + B[0]) / 2., (A[1] + B[1]) / 2.)
                        verts.extend([C, B])
                        codes.extend([Path.CURVE3, Path.CURVE3])
                    verts.append(segment[-1])
                    codes.append(Path.CURVE3)

            start = end + 1

        path = Path(verts, codes)
        self.cache[char] = path
        glyph = PathPatch(path, antialiased=True)

        return glyph




from __future__ import division, print_function

from functools import partial
from glob import glob
from hashlib import md5
from multiprocessing import Pool
from os import mkdir
from os.path import abspath, dirname, exists, join
from textwrap import dedent
from zipfile import is_zipfile, ZipFile

import bottle

from bottle import get, post, redirect, request, route, static_file


pool = Pool(processes=3)
views = join(dirname(abspath(__file__)), 'views')

bottle.TEMPLATE_PATH.insert(0, views)


def process_job(seqfile):
    pass

@get('/')
def form():
    return static_file('form.html', root=views)

@post('/')
def submit():
    data = request.files.data
    m = md5()
    if not is_zipfile(data.file):
        return static_file('invalid.html', root=views)
    data.file.seek(0)
    while True:
        p = data.file.read(8192)
        if not p:
            break
        m.update(p)
    while True:
        if not exists(m.hexdigest()):
            break
        m.update(m.digest())
    jid = m.hexdigest()
    data.file.seek(0)
    zipdata = ZipFile(data.file)
    mkdir(jid)
    zipdata.extractall(jid)
    seqfiles = glob(join(jid, '*.fna')) + glob(join(jid, '*.fasta')) + glob(join(jid, '*.fa'))
    if len(seqfiles) != 1:
            return static_file('invalid.html', root=views)
    pool.apply_async(process_job, (seqfiles[0],), callback=fill_job_templates, error_callback=partial(report_job_error, jid=jid))
    redirect('/job/%s/' % jid)

@route('/job/<jid>/<part>')
def job(jid=None, part='index'):
    if jid is None or not exists(jid):
        return static_file('404.html', root=views)
    elif not exists(join(jid, 'index.html')):
        return static_file('unfinished.html', root=views)
    else:
        return static_file(join(jid, part + '.html'))


parts = set( 'index'
           , 'download'
           , 'coverage'
           , 'majority'
           , 'coverage+majority'
           , 'maxdivergence'
           , 'sitewise'
           )

def partfile(jid, part):
    return join(jid, part + '.html')

def report_job_error(jid, exc):
    with open(join(jid, 'index.html'), 'w') as fh:
        fh.write('error occurred: %s' % str(exc))

def fill_job_templates(result):
    jid = result.get()
    with open(join(jid, jid + '.json')) as data:
        for part in parts:
            file = partfile(jid, part)
            if not exists(file):
                with open(file, 'w') as fh:
                    if part == 'index':
                        fh.write(dedent('''\
                        filtering of reads (downloads, coverage, majority, cov+maj
                        maximum divergence
                        sitewise mutation rate classes'''))
                    if part == 'download':
                        fh.write('treat this as a download of the filtered sequences')
                    if part == 'coverage':
                        fh.write('treat this as a download of the coverage pdf')
                    if part == 'majority':
                        fh.write('treat this as a download of the majority pdf')
                    if part == 'coverage+majority':
                        fh.write('treat this as a download of the coverage+majority pdf')
                    if part == 'maxdivergence':
                        fh.write('1 index for maximum divergence with 2 for each window: alignment and tree and 3 for the max window: pdf tree')
                    if part == 'sitewise':
                        fh.write('2 files, html and csv file, with 1 profile index for each site')




from __future__ import division, print_function

import json

from collections import Counter, namedtuple
from math import ceil, log
from os import close, remove
from os.path import abspath, exists, join, split
from sys import stderr
from tempfile import mkstemp
from textwrap import dedent

from Bio.Seq import translate
from Bio.Alphabet import generic_nucleotide

from hppy import HyphyInterface, HyphyMap


__all__ = ['DiversityEstimator', 'Thresholds']


class Thresholds(object):

    def __init__(self, alnscore, mincopy=None, minprop=None):
        if mincopy == None and minprop == None:
            raise ValueError('one of either mincopy or minprop must be provided')
        self.alnscore = alnscore
        self.mincopy = 0 if mincopy is None else mincopy
        self.minprop = 0 if minprop is None else minprop


class DiversityEstimator(HyphyMap):

    def __init__(self, batchfile=None, retvar=None):
        if batchfile is None:
            batchfile = join(
                    split(abspath(__file__))[0],
                    'hyphy', 'divestimator.bf'
            )

        if retvar is None:
            retvar = "_divest_outstr"

        if not exists(batchfile):
            raise ValueError("Invalid batchfile `%s', it doesn't exist!" % batchfile)
        super(DiversityEstimator, self).__init__(batchfile, retvar)

    def __call__(self, refseq, msa, scores, thresholds, quiet=True):
        return DiversityEstimator.estimate(self, refseq, msa, scores, quiet)

    def estimate(self, refseq, msa, scores, thresholds, quiet=True):
        if not isinstance(thresholds, Thresholds):
            raise TypeError("unsupported operand type(s) for estimate: '%s'" % str(type(thresholds)))

        # assume MSA has been trimmed to A:B
        # bin unique variants
        # strip all sequences that don't have minimum scores
        nuc_bins = Counter(str(r.seq) for r, s in zip(msa, scores) if s > thresholds.alnscore)
        amino_bins = Counter()
        for cdn, c in nuc_bins.items():
            amino_bins[translate(cdn)] += c

        numvariants = sum(nuc_bins.values())

        # only keep those that have the minimum copy count
        mincopy = max(thresholds.mincopy, thresholds.minprop * numvariants)
        retained = [r for r, c in amino_bins if c > mincopy]

        inputfile, treefile, bootrawfile, bspsfile = [None] * 4
        try:
            fd, inputfile = mkstemp(); close(fd)
            fd, treefile = mkstemp(); close(fd)
            fd, bootrawfile = mkstemp(); close(fd)
            fd, bspsfile = mkstemp(); close(fd)

            with open(inputfile, 'w') as fh:
                print('\n'.join('>%d\n%s' % (i, seq) for i, seq in enumerate(retained)), file=fh)

            iface = HyphyInterface()
            iface.runqueue(execstr=dedent('''\
            _divest_stdlib = HYPHY_LIB_DIRECTORY + "TemplateBatchFiles" + DIRECTORY_SEPARATOR;
    
            _divest_njopts = {};
            _divest_njopts[ 0 ] = "Distance formulae";
            _divest_njopts[ 1 ] = "Nucleotide/Protein";
            _divest_njopts[ 2 ] = "%(input_file)s";
            _divest_njotps[ 3 ] = "Force Zero";
            _divest_njopts[ 4 ] = "TN93";
            _divest_njopts[ 5 ] = "y";
            _divest_njopts[ 6 ] = "%(tree_file)s";
            ExecuteAFile( _divest_stdlib + "NeighborJoining.bf", _divest_njopts );
    
            _divest_bsopts = {};
            _divest_bsopts[ 0 ] = "100";
            _divest_bsopts[ 1 ] = "%(bootraw_file)s";
            _divest_bsopts[ 2 ] = "Proportions";
            _divest_bsopts[ 3 ] = "70";
            _divest_bsopts[ 4 ] = "%(bootstrap_ps_file)s";
            ExecuteAFile( _divest_stdlib + "post_npbs.bf", _divest_bsopts );
    
            _divest_anpdopts = {};
            _divest_anpdopts[ 0 ] = "%(input_file)s";
            _divest_anpdopts[ 1 ] = "HKY85";
            _divest_anpdopts[ 2 ] = "Global";
            _divest_anpdopts[ 3 ] = "%(tree_file)s";
            ExecuteAFile( _divest_stdlib + "AnalyzeNucProtData.bf", _divest_anpdopts );
            ''') % {
                'input_file': inputfile,
                'tree_file': treefile,
                'bootraw_file': bootrawfile,
                'bootstrap_ps_file': bspsfile
            })

            # self.map()
        finally:
            if exists(inputfile):
                remove(inputfile)
            if exists(treefile):
                remove(treefile)
            if exists(bootrawfile):
                remove(bootrawfile)
            if exists(bspsfile):
                remove(bspsfile)

        # define window A, B
        # align sequences to ref, truncating to [A, B)
        # create bins for each obs in window by aa, and codons
        # the total # of unique variants == len(binner)
        # filter variants by count or % threshold (* len(sequences))
        # do NJ with "Distance formulae", "Nucleotide/protein", inFile, "Force Zero", "TN93", "y", inFile + ".tree"
        #   @ HYPHY_LIB_DIRECTORY + "TemplateBatchFiles" + DIRECTORY_SEPARATOR + "NeighborJoining.bf"
        # do bootstrap with "100", inFile + ".bootraw", "Proportions", "70", inFile + ".njboot.ps"
        #   @ HYPHY_LIB_DIRECTORY + "TemplateBatchFiles" + DIRECTORY_SEPARATOR + "post_npbs.bf"
        # do NucProt analysis with inFile, "HKY85", "Global", inFile + ".tree"
        #   @ HYPHY_LIB_DIRECTORY + "TemplateBatchFiles" + DIRECTORY_SEPARATOR + "AnalyzeNucProtData.bf" 
        # for each replicate (Map part):
        #   do NJ like above with 2: inFile + ".sim" and 6: inFile + ".sim.tree", and "nj" in EAB
        #   do NucProt analysis like above with 0: inFile + ".sim" and 3: inFile + ".sim.tree", and "sim" in EAB
        #   do simD[replicate] = maxPathInATree( "sim.givenTree" ); 




from __future__ import division, print_function

import json

from math import ceil, log
from os.path import abspath, exists, join, split
from sys import stderr

from Bio.Alphabet import generic_nucleotide

from hppy import HyphyMap


__all__ = ['CodonAligner']


class CodonAligner(HyphyMap):

    def __init__(self, batchfile=None, retvar=None):
        if batchfile is None:
            batchfile = join(
                    split(abspath(__file__))[0],
                    'hyphy', 'codonaligner.bf'
            )

        if retvar is None:
            retvar = "_cdnaln_outstr"

        if not exists(batchfile):
            raise ValueError("Invalid batchfile `%s', it doesn't exist!" % batchfile)
        super(CodonAligner, self).__init__(batchfile, retvar)

    def __call__(self, refseq, seqs, revcomp=False, expected_identity=0., keep_insertions=False, quiet=True):
        return CodonAligner.align(self, refseq, seqs, revcomp, expected_identity, keep_insertions, quiet)

    def align(self, refseq, seqs, revcomp=False, expected_identity=0., keep_insertions=False, quiet=True):
        # if we have no sequences, abort early to prevent later errors
        if not len(seqs):
            return [], []

        # uppercase the refseq to deal with bugs in HyPhy's codon aligner
        refseq = refseq.upper()

        numseqs = len(seqs)
        # if the # nodes exceeds the number of seqs, we just need numseqs jobs
        numnodes = min(numseqs, self.nodes)
        seqs_per_node = max(1, numseqs // numnodes)
        remainder = numseqs % numnodes

        arg1 = 'Yes' if revcomp else 'No'
        arg2 = 'Yes' if keep_insertions else 'No'

        argslist = []
        lwr, upr = 0, 0

        for i in range(numnodes):
            # since our traversal is stateful, keep these cursors
            # around. During the first remainder iterations,
            # add an extra seq to the list of seqs, afterwards
            # proceed as normal 
            lwr = upr
            if i < remainder:
                upr = min(numseqs, lwr + seqs_per_node + 1)
            else:
                upr = min(numseqs, lwr + seqs_per_node)
            node_seqs = [s.upper() for s in seqs[lwr:upr]]
            argslist.append( [arg1, arg2, refseq, expected_identity, len(node_seqs)] + node_seqs )

        retstrs = self.map(argslist, quiet)

        seqscores = []
        for retstr in retstrs:
            seqscores.extend(json.loads(retstr))

        newseqstrs, scores, overlaps, identities = zip(*seqscores)

        return list(newseqstrs), list(scores), list(overlaps), list(identities)




from __future__ import division, print_function

import matplotlib as mpl
mpl.use('pdf')

import matplotlib.pyplot as plt
import numpy as np

from collections import defaultdict, Counter
from itertools import repeat
from operator import itemgetter
from os import close
from os.path import dirname, join
from re import compile as re_compile
from tempfile import mkstemp

from Bio.Alphabet import Gapped, HasStopCodon, _verify_alphabet
from Bio.Alphabet.IUPAC import (ambiguous_dna, ambiguous_rna,
        extended_protein, unambiguous_dna, unambiguous_rna)
from Bio.Motif import Motif
from Bio.Seq import Seq

from matplotlib.font_manager import (createFontList,
        findSystemFonts, fontManager)
from matplotlib.lines import Line2D
from matplotlib.patches import Rectangle
from matplotlib.ticker import FormatStrFormatter, FuncFormatter
from matplotlib.transforms import Affine2D

from BioExt import _GAP, _STOP

from ._basefont import Basefont


__all__ = [
    'COVERAGE',
    'MAJORITY',
    'graph_coverage_majority',
    'graph_logo'
]


_HY454_FONT_PATHS = [join(dirname(__file__), 'data', 'fonts', 'ttf')]


# update the fontManager to handle the Roboto font installed with hy454
fontManager.ttffiles.extend(findSystemFonts(_HY454_FONT_PATHS))
fontManager.ttflist = createFontList(fontManager.ttffiles)


def _max_nonzero_min(values, default=0):
    vmax = max(values)
    vmin = vmax
    for v in values:
        if v > 0 and v < vmin:
            vmin = v
    vmin = default if vmin == vmax else vmin
    return vmax, vmin


COVERAGE, MAJORITY = 1, 2

def graph_coverage_majority(
    alignment,
    mode,
    filename=None,
    dpi=None, figsize=None, format='pdf', transparent=True
):
    if not mode:
        mode = COVERAGE | MAJORITY

    if mode not in (COVERAGE, MAJORITY, COVERAGE | MAJORITY):
        raise ValueError('mode requires COVERAGE, MAJORITY, or COVERAGE | MAJORITY')

    if filename is None:
        fd, filename = mkstemp(); close(fd)

    if figsize is None:
        # actually results in a graph that is almost 4" high
        # and 6.4" wide -- almost the golden ratio
        figsize = (6, 6)

    n0 = 0
    M = len(alignment)
    N = alignment.get_alignment_length()

    # if we don't match the whole reference,
    # cut off the head, tail of gaps
    leading = re_compile(r'^-*')
    trailing = re_compile(r'-*$')
    l = -N
    n = -N
    for r in alignment:
        seq = str(r.seq)
        # *ing.search should always match
        v1 = len(leading.search(seq).group(0))
        v2 = len(trailing.search(seq).group(0))
        if v1 < abs(l):
            l = v1
        if v2 < abs(n):
            n = v2
    # if we're greater than 0, remove leading, trailing gaps
    if l > 0:
        n0 += l
    if n > 0:
        N -= n

    frac = 1. / M
    heights = np.zeros((N - n0,), dtype=float)

    xs = np.arange(n0 + 1, N + 1)

    # some heuristic to prevent too many ticks,
    # is complicated by the start and end dynamically business
    # 0.477122 jumps an order 10 at 30% of the next order 10
    specialK = 0.477122
    xdiv = 10 ** int(np.log10(N) - specialK) or 1
    xsep = xdiv * int((N - n0) / 5 / xdiv + 1)
    xstart = (n0 // xsep) * xsep
    xend = N + xsep
    xticks = np.arange(xstart, xend, xsep)
    if len(xticks) > 1:
        fix = False
        delta0 = xticks[1] - n0
        deltaN = N - xticks[-2]
        min_delta = xsep / 4
        if delta0 != 0 and delta0 < min_delta:
            xstart += xsep
            fix = True
        if deltaN != 0 and deltaN < min_delta:
            xend -= xsep
            fix = True
        if fix:
            xticks = np.arange(xstart, xend, xsep)
    xticks[0] = n0 + 1
    xticks[-1] = N

    ydiv = 10 ** int(np.log10(M) - specialK) or 1
    ysep = ydiv * int(M / 5 / ydiv + 1)
    yticks = np.arange(ysep, M+ysep, ysep)
    if len(yticks) > 1:
        if M - yticks[-2] < ysep / 4:
            yticks = np.arange(ysep, M, ysep)
    yticks[-1] = M

    mpl.rcParams['font.family'] = 'sans-serif'
    mpl.rcParams['font.sans-serif'] = 'Roboto'

    fig = plt.figure(figsize=figsize, dpi=dpi)

    # golden rectangle! 
    rect = 0.2, 0.2, 1, 0.618
    ax1 = fig.add_axes(rect)

    majorities = np.zeros((N - n0,), dtype=float)
    if mode & MAJORITY:
        for i, col in enumerate(range(n0, N)):
            # count the number of occurrences of each unique character
            # then grab the maximum count observed and multiply it by
            # frac to get the majorities
            counts = Counter(alignment[:, col].upper())
            if _GAP in counts:
                del counts[_GAP]
            # grab the count of the most common variant
            m = 0
            if len(counts):
                m = counts.most_common(1)[0][1]
            # compute the normalized majority in the majority-only case,
            # taking care to avoid division-by-zero errors
            if mode == MAJORITY:
                s = sum(counts.values())
                s = 1 if s == 0 else s
                majorities[i] = m / s
            else:
                majorities[i] = m * frac
        ax1.plot(
            xs, majorities,
            color=_LRED, linewidth=1., zorder=-1
        )

    if mode & COVERAGE:
        for i, col in enumerate(range(n0, N)):
            heights[i] = sum(frac for p in alignment[:, col] if p != _GAP)

    # labels
    ax1.set_xlabel('Reference sequence position')

    extra_artists = []

    if mode == (COVERAGE | MAJORITY):
        ax1.plot(
            xs, heights,
            color=_LBLUE, linewidth=1., zorder=-2
        )
        # create a proxy artist for legend, PolyCollections don't work (heights)
        p1 = Line2D([0, 1], [0, 1], color=_LBLUE, linewidth=1.)
        # create a proxy artist for legend, [Lines2D] don't work (majorities)
        p2 = Line2D([0, 1], [0, 1], color=_LRED, linewidth=1.)
        leg = ax1.legend(
            [p1, p2], ['Coverage', 'Majority'],
            bbox_to_anchor=(0.5, -0.15), loc=9, ncol=2,
            prop={ 'size': 12 }, borderpad=0.
        )
        leg.legendPatch.set_alpha(0.)
        extra_artists.append(leg)
#         ax1.set_ylabel('Coverage - majority')
    elif mode == COVERAGE:
        ax1.fill_between(
            xs, heights, majorities,
            edgecolor=_LBLUE, facecolor=_LBLUE, linewidth=1., zorder=-1
        )
        ax1.set_ylabel('Coverage')
    else:
        ax1.set_ylabel('Proportion')

    def format_percent(x, pos=None):
        return '%1.0f%%' % (100 * x)

    ax1.yaxis.set_major_formatter(FuncFormatter(format_percent))
    ax1.set_yticks(np.arange(0.2, 1.1, 0.2))
    ax1.set_xlim((n0 + 1, N))

    major_ticks = ax1.xaxis.get_major_ticks(len(xticks))

    # disable the first and last tick on the x-axis,
    # they're redundant and ugly (esp if minH > 0)
    major_ticks[0].tick1On = False
    major_ticks[-1].tick1On = False

    # if we're only doing coverage, include the number
    # of sequences, but with majority in the mix
    # it doesn't make sense 
    if mode == MAJORITY:
        ax1.set_xticks(xticks)
        # if we're not showing the # of sequences on the left,
        # remove the ticks and spine
        major_ticks += ax1.yaxis.get_major_ticks()
        ax1.spines['right'].set_visible(False)
        # alter axes to show max, min value
        maxM, minM = _max_nonzero_min(majorities)
        ax1.spines['left'].set_bounds(minM, maxM)
    else:
        ax2 = ax1.twinx()
        ax2.set_ylabel('No. of sequences', rotation=270.)
        ax2.set_xticks(xticks)
        ax2.set_yticks(yticks)
        # set transparent here otherwise ax2 doesn't exist
        if transparent:
            ax2.patch.set_alpha(0.)
        # get the major ticks so we can disable them later
        major_ticks += ax2.xaxis.get_major_ticks()
        # disable the top spines, like we do later
        ax2.spines['top'].set_visible(False)
        # use the axes spines to show the maximum value
        maxH, minH = _max_nonzero_min(heights)
        ax1.spines['left'].set_bounds(minH, maxH)
        ax1.spines['right'].set_bounds(minH, maxH)

    if transparent:
        fig.patch.set_alpha(0.)
        ax1.patch.set_alpha(0.)

    # remove the upper ticks 
    for tick in major_ticks:
        tick.tick2On = False

    # remove the upper axis border
    ax1.spines['top'].set_visible(False)

    fig.savefig(
        filename,
        format=format,
        transparent=transparent,
        bbox_extra_artists=extra_artists,
        bbox_inches='tight',
        pad_inches=0.25
    )

    return filename


_DNA_ALPHABET = Gapped(ambiguous_dna)
_RNA_ALPHABET = Gapped(ambiguous_rna)
_AMINO_ALPHABET = HasStopCodon(Gapped(extended_protein, gap_char=_GAP), stop_symbol=_STOP)

def _fix_ambigs(pwm, alphabet):
    mapper = {}
    killchars = _GAP
    # killchars ambigs by distributing their probability uniformly 
    if alphabet == _DNA_ALPHABET or alphabet == _RNA_ALPHABET:
        T = 'T' if alphabet == _DNA_ALPHABET else 'U'
        mapper.update({
            'M': 'AC',
            'R': 'AG',
            'W': 'A' + T,
            'S': 'CG',
            'Y': 'C' + T,
            'K': 'G' + T,
            'V': 'ACG',
            'H': 'AC' + T,
            'D': 'AG' + T,
            'B': 'CG' + T,
            # 'N': 'ACG' + T
        })
        killchars += 'N'
    elif alphabet == _AMINO_ALPHABET:
        mapper.update({
            'B': 'DN',
            'J': 'IL',
            'Z': 'EQ',
            # 'X': 'ACDEFGHIKLMNPQRSTVWY',
        })
        killchars += _STOP + 'X' + 'OU'
    for i in range(len(pwm)):
        for k, unambig in mapper.items():
            if k in pwm[i]:
                # redistribute the probability uniformly
                C = pwm[i][k] / len(unambig)
                for l in unambig:
                    pwm[i][l] += C
                # killchars the key
                del pwm[i][k]
    # killchars gaps
    for i in range(len(pwm)):
        for char in killchars:
            if char in pwm[i]:
                # uniform redistribution to everybody else, eg p(letter | observed)
                C = pwm[i][char] / (len(pwm[i])-1)
                for l in pwm[i].keys():
                    pwm[i][l] += C
                # remove the gap
                del pwm[i][char]
    return pwm


_BLACK = '#000000'
_GREY = '#969cb0'
_GREEN = '#b8ff25'
_ORANGE = '#ff6e27'
_BLUE = '#189cff'
_RED = '#e80c5b' # '#e80c7a'

# android color swatch
_LGREY = '#F2F2F2'
_LBLUE = '#33B5E5'
_LPURPLE = '#AA66CC'
_LGREEN = '#99CC00'
_LORANGE = '#FFBB33'
_LRED = '#FF4444'
_DGREY = '#DDDDDD'
_DBLUE = '#0099CC'
_DPURPLE = '#9933CC'
_DGREEN = '#669900'
_DORANGE = '#FF8800'
_DRED = '#CC0000'

# default grey
_DNA_COLORS = defaultdict(repeat(_GREY).__next__, {
    'A': _LGREEN,
    'C': _LBLUE,
    'G': _LORANGE,
    'T': _LRED,
    'U': _LRED
})
_AMINO_COLORS = defaultdict(repeat(_GREY).__next__,
    [(l, _LGREEN ) for l in 'KRH'] + \
    [(l, _LBLUE  ) for l in 'DE'] + \
    # [(l, _ORANGE) for l in ''] + \
    [(l, _LRED   ) for l in 'AVLIPWFM']
)


def graph_logo(
    alignment,
    columns,
    filename=None,
    dpi=None, edgecolor='k', figsize=None, format='pdf', labels=None, linewidth=0., transparent=True
):
    if filename is None:
        fd, filename = mkstemp(); close(fd)

    if figsize is None:
        figsize = (3, 3)

    if labels is None:
        labels = ['%d' % (idx + 1) for idx in columns]

    M = len(alignment)
    N = len(columns)

    alph = None
    for _alph in (_DNA_ALPHABET, _RNA_ALPHABET, _AMINO_ALPHABET):
        for r in alignment:
            r.seq.alphabet = _alph
        if all([_verify_alphabet(r.seq.upper()) for r in alignment]):
            alph = _alph
            break
    if alph is None:
        raise RuntimeError("sequences with indeterminable alphabet provided")

    motif = Motif(alphabet=alph)

    instances = [''.join(z).upper() for z in zip(*[alignment[:, i] for i in columns])]
    for instance in instances:
        motif.add_instance(Seq(instance, alph))

    # set laplace = True to include the backgrounds
    pwm = _fix_ambigs(motif.pwm(laplace=False), alph)

    # heuristic to determine whether nucleotide or protein alphabet
    # need to use either base 4 or 20 depending 
    alphlen, _alphkeys = max([(len(pwm[i]), pwm[i].keys()) for i in range(N)], key=itemgetter(0))
    s, colors = (4, _DNA_COLORS) if alphlen < 20 else (20, _AMINO_COLORS)
    alphkeys = ['']
    alphkeys.extend(_alphkeys)
    alphmap = dict(zip(alphkeys, range(len(alphkeys))))

    # compute the information content at each position 
    maxbits = np.log2(s)
    e_n = (s - 1) / (2. * np.log(2) * M)
    R = maxbits * np.ones((N,), dtype=float)
    R -= [-sum([v * np.log2(v) for _, v in pwm[i].items() if v > 0.]) for i in range(N)]
    R -= e_n

    heights = np.zeros((alphlen, N), dtype=float)
    identities = np.zeros((alphlen, N), dtype=int)

    for j in range(N):
        i = 0
        for k, v in sorted(pwm[j].items(), key=itemgetter(1)):
            heights[i, j] = R[j] * v
            identities[i, j] = alphmap[k]
            i += 1

    font = Basefont(join(_HY454_FONT_PATHS[0], 'Roboto-Black.ttf'))

    mpl.rcParams['font.family'] = 'sans-serif'
    mpl.rcParams['font.sans-serif'] = 'Roboto'

    fig = plt.figure(figsize=figsize, dpi=dpi)

    # make each column a vertical golden rect 
    rect = 0.2, 0.2, 0.382 * N, 0.618
    ax = fig.add_axes(rect)

    ax.set_ylabel('bits')

    if figsize is None:
        fig.set_figwidth(N)

    if transparent:
        fig.patch.set_alpha(0.)
        ax.patch.set_alpha(0.)

    # remove the top and right ticks
    for tick in ax.xaxis.get_major_ticks() + ax.yaxis.get_major_ticks():
        tick.tick2On = False

    # remove the bottom ticks
    for tick in ax.xaxis.get_major_ticks():
        tick.tick1On = False

    # rotate the x-axis labels by 45 degrees to enhance packing
    for label in ax.xaxis.get_ticklabels():
        label.set_rotation(45)

    # disable top and right spines, we don't need them
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    def format_xlabel(x, pos=None):
        idx = np.clip(int(x)-1, 0, N-1)
        return labels[idx]

    ax.xaxis.set_major_formatter(FuncFormatter(format_xlabel))
    # avoid too much precision
    ax.yaxis.set_major_formatter(FormatStrFormatter('%1.1f'))

    # set the ticks
    ax.set_yticks(np.append(np.arange(0, maxbits, 0.5, dtype=float), maxbits))
    ax.set_xticks(np.arange(1, N+1, dtype=float) + 0.5)

    # set the axes limits here AFTER the ticks, otherwise borkage
    ax.set_xlim((1, N+1))
    ax.set_ylim((0, maxbits))

    idxs = np.arange(1, N+1)
    bottoms = np.zeros((N,), dtype=float)
    for i in range(alphlen):
        bars = ax.bar(idxs, heights[i, :], width=1., bottom=bottoms)
        bottoms += heights[i, :]
        for j, bar in enumerate(bars):
            if identities[i, j]:
                l = alphkeys[identities[i, j]]
                glyph = font[l]
                ax.add_patch(glyph)
                glyph.set_transform(bar.get_transform())
                bar.set_visible(False)
                glyph.set_edgecolor(edgecolor)
                glyph.set_facecolor(colors[l])
                glyph.set_linewidth(linewidth)
                glyph.set_zorder(-1)

    # set the remaining spine to show the maximum value
    ax.spines['left'].set_bounds(0., max(bottoms))

    fig.savefig(filename, format=format, transparent=transparent, bbox_inches='tight', pad_inches=0.25)

    return filename



#!/usr/bin/env python

from __future__ import division, print_function

import sys

from os.path import abspath, join, split
from setuptools import setup

sys.path.insert(0, join(split(abspath(__file__))[0], 'lib'))
from hy454 import __version__ as _hy454_version

setup(name='hy454',
      version=_hy454_version,
      description='454 UDS analysis tools',
      author='N Lance Hepler',
      author_email='nlhepler@gmail.com',
      url='http://github.com/veg/hy454',
      license='GNU GPL version 3',
      packages=['hy454'],
      package_dir={'hy454': 'lib/hy454'},
      package_data={'hy454': [
            'hyphy/*.bf',
            'data/fonts/ttf/*.ttf',
            'data/scores/*.dat'
    ]},
      scripts=[
            'scripts/codonaligner',
            'scripts/coveragegrapher',
            'scripts/seqlogo'
      ],
      requires=['Bio', 'BioExt', 'fakemp', 'freetype', 'hppy', 'matplotlib', 'numpy']
     )




from __future__ import division, print_function

from copy import deepcopy
from math import ceil, log
from multiprocessing import cpu_count, current_process
from operator import itemgetter
from re import compile as re_compile, I as re_I
from sys import exc_info, exit as sys_exit, float_info

from Bio import SeqIO
from Bio.Align import MultipleSeqAlignment
from Bio.Alphabet import generic_nucleotide
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord

from BioExt import _GAP, enumerate_by_codon

from fakemp import farmout, farmworker

from ._codonaligner import CodonAligner


__all__ = [
    'preprocess_seqrecords',
    'CUSTOM', 'FIRST', 'LONGEST',
    'determine_refseq',
    'align_to_refseq',
    'from_positional',
    'to_positional'
]


def preprocess_seqrecords(seqrecords):
    remove_unknown = re_compile(r'[^ACGTUWSMKRYBDHVN]', re_I)
    strip_front = re_compile(r'^[N]+', re_I)
    strip_rear = re_compile(r'[N]+$', re_I)

    for record in seqrecords:
        seq = str(record.seq)
        seq = remove_unknown.sub('', seq)
        seq = strip_front.sub('', seq)
        seq = strip_rear.sub('', seq)

        record.seq = Seq(seq, generic_nucleotide)

    return


CUSTOM, FIRST, LONGEST = range(3)
def determine_refseq(seqrecords, mode):
    if mode not in range(3):
        raise ValueError("mode must be one of CUSTOM, FIRST, or LONGEST")

    if mode == CUSTOM:
        seq = input("Input the entire reference sequence without newlines :: ")
        refseq = SeqRecord(Seq(seq, generic_nucleotide),
                id="ref", name="reference",
                description="Custom reference sequence")
    elif mode == FIRST:
        refseq = seqrecords.pop(0)
    elif mode == LONGEST:
        idx, refseq = max(enumerate(seqrecords), key=lambda r: len(r[1].seq))
        seqrecords.pop(idx)

    return refseq, seqrecords


def align_to_refseq(refseq, seqrecords, revcomp=True, expected_identity=0., keep_insertions=False, quiet=False):
    aligned, _, _, identities = CodonAligner()(
        str(refseq.seq),
        [str(s.seq) for s in seqrecords],
        revcomp,
        expected_identity,
        keep_insertions,
        quiet
    )

    # deepcopy the seqrecords so that we can change their sequences later
    aligned_records = []
    discarded_records = []
    for i, aln in enumerate(aligned):
        old = seqrecords[i]
        if expected_identity > 0. and identities[i] < 0:
            discarded_records.append(old)
        else:
            new = SeqRecord(
                Seq(aln, generic_nucleotide),
                old.id,
                old.name,
                old.description,
                deepcopy(old.dbxrefs),
                deepcopy(old.features),
                deepcopy(old.annotations)
                # don't grab the letter_annotations,
                # they won't match anymore
            )
            aligned_records.append(new)

    if not keep_insertions:
        return MultipleSeqAlignment(aligned_records), discarded_records

    return aligned_records, discarded_records


def from_positional(datastruct):
    records = []
    gapcdn = _GAP * 3
    maxlen = 0
    for poscdns in datastruct.values():
        lastcdn, _ = poscdns[-1]
        if lastcdn > maxlen:
            maxlen = lastcdn
    maxlen += 3
    for seqid, poscdns in datastruct.items():
        seqstr = ''
        prev = 0
        for pos, cdn in poscdns:
            seqstr += _GAP * (pos - prev) + cdn
            prev = pos + 3
        seqstr += _GAP * (maxlen - prev)
        seq = Seq(seqstr, generic_nucleotide)
        record = SeqRecord(
            seq,
            id=seqid,
            name=seqid,
            description=seqid
        )
        records.append(record)
    return MultipleSeqAlignment(records)


def to_positional(msa):
    datastruct = {}
    gapcdn = _GAP * 3
    for seq in msa:
        seqdata = []
        for pos, cdn in enumerate_by_codon(seq):
            if cdn != gapcdn:
                seqdata.append((pos, cdn))
        if len(seqdata):
            datastruct[seq.id] = seqdata
    return datastruct
